
/*
 Copyright 2009 Eigenlabs Ltd.  http://www.eigenlabs.com

 This file is part of EigenD.

 EigenD is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 EigenD is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with EigenD.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SIMPLEAGENT__
#define __SIMPLEAGENT__

#include <piw/piw_bundle.h>
#include <piw/piw_data.h>
#include <piw/piw_clock.h>
#include <piw/piw_tsd.h>
#include <picross/pic_ilist.h>
#include <picross/pic_weak.h>

class simpleagent_wire_t;

struct simpleagent_impl_t: piw::root_t, piw::root_ctl_t, piw::clocksink_t, virtual pic::lckobject_t
{
	simpleagent_impl_t(piw::clockdomain_ctl_t *cd, const piw::cookie_t &c, const char* name);

    virtual ~simpleagent_impl_t();

    void clocksink_ticked(unsigned long long f, unsigned long long t);

    void add_ticker(simpleagent_wire_t *w);

    void del_ticker(simpleagent_wire_t *w);

    void invalidate();

    virtual simpleagent_wire_t* create_wire(const piw::event_data_source_t& es) = 0;

    piw::wire_t *root_wire(const piw::event_data_source_t &es);

    void root_closed();

    void root_opened();

    void root_clock();

    void root_latency();

    pic::lckmap_t<piw::data_t, simpleagent_wire_t *>::lcktype children_;
    pic::ilist_t<simpleagent_wire_t> tickers_;
    bct_clocksink_t *up_;
    pic::weak_t<simpleagent_wire_t> current_wire_;
};

/*
 * Static method to be called by the fast thread.
 */
static int __wire_invalidator(void *w_, void *_);

struct simpleagent_wire_t
   : piw::wire_t, piw::wire_ctl_t, piw::event_data_sink_t,
     piw::event_data_source_real_t, virtual pic::lckobject_t,
     pic::element_t<>, pic::tracked_t
{
protected: // prevent allocation on stack to allow self destruction in wire_closed
	simpleagent_wire_t(simpleagent_impl_t *p, const piw::event_data_source_t& es, unsigned int num_out_signals);
public:
	virtual ~simpleagent_wire_t();

	void wire_closed();

	void invalidate();

	void event_start(unsigned seq, const piw::data_nb_t &id, const piw::xevent_data_buffer_t &b);

	bool event_end(unsigned long long t);

	void event_buffer_reset(unsigned s, unsigned long long t, const piw::dataqueue_t &o, const piw::dataqueue_t &n);

	virtual void process(unsigned, const piw::data_nb_t &, unsigned long long) = 0;

	void ticked(unsigned long long f, unsigned long long t);

	void source_ended(unsigned seq);

	simpleagent_impl_t *root_;
	piw::xevent_data_buffer_t::iter_t input_;
	piw::xevent_data_buffer_t output_;

	unsigned int num_out_signals_;

	unsigned long long last_from_;
};

// Convenience Macros to declare and define the derived agent_t::impl and agent_wire_t classes for a concrete simple agent
// in order to help to keep focus on the main purpose of the agent.
// Can of course also be done manually if a macro free style is preferred over conciseness :)

#define DECLARE_FASTCALL_PROPERTY(property_t, property) \
	property_t property##_; \
	static int set_##property##_static(void* i, void* par); \
    void set_##property(const property_t& property);

#define DECLARE_FASTCALL_PROPERTY_MAP(key_t, key, value_t, value) \
	std::map<key_t,value_t> key##_map_; \
	static int set_##key##_static(void* i, void* par); \
    void set_##key(const key_t& key, const value_t& value);

#define DEFINE_FASTCALL_PROPERTY(agent,property_t,property) \
	int agent##_t::impl_t::set_##property##_static(void* i, void* par) \
	{ \
		agent##_t::impl_t *i_ = (agent##_t::impl_t *)i; \
		int par_ = *(property_t *)par; \
		i_->set_##property(par_); \
		return 0; \
	} \
	void agent##_t::set_##property(property_t val) \
	{ \
		piw::tsd_fastcall(agent##_t::impl_t::set_##property##_static, impl_, &val); \
	}

#define DEFINE_FASTCALL_PROPERTY_MAP(agent,key_t,key,value_t,value) \
	int agent##_t::impl_t::set_##key##_static(void* i, void* par) \
	{ \
		agent##_t::impl_t *i_ = (agent##_t::impl_t *)i; \
		std::pair<key_t,value_t> par_ = *(std::pair<key_t,value_t>*)par; \
		i_->set_##key##(par_.first, par_.second); \
		return 0; \
	} \
	void agent##_t::set_##key(key_t key, value_t value) \
	{ \
		std::pair<key_t, value_t> tuple(key, value); \
		piw::tsd_fastcall(agent##_t::impl_t::set_##key##_static, impl_, &tuple); \
	}

#endif // __SIMPLEAGENT__

